﻿using System;
using System.Collections.Generic;
using System.Drawing;
using System.IO;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

namespace Jewles
{
     sealed class GameField
     {
          public enum Gems : byte
          {
               None = 0,
               Red = 1,
               Green = 2,
               Blue = 3,
               White = 4,
               Yellow = 5,
               Violet = 6
          }

          public readonly Size Size = new Size(Constants.FIELD_WIDTH, Constants.FIELD_HEIGHT);
          private readonly Gems[,] _field = new Gems[Constants.FIELD_WIDTH, Constants.FIELD_HEIGHT];
          private readonly GemFactory _gf;

          public GameField()
          {
               do
               {
                    _gf = new GemFactory();
                    for (int i = 0; i < Constants.FIELD_WIDTH; i++)
                         for (int j = 0; j < Constants.FIELD_HEIGHT; j++)
                              _field[i, j] = _gf.GetGem(i);
                    List<GemGroup> groups = new List<GemGroup>(FindGroups());
                    do
                    {
                         if (groups.Count > 0)
                              foreach (GemGroup g in groups)
                                   foreach (Point p in g.Group)
                                        _field[p.X, p.Y] = _gf.GetGem(p.X);

                         groups.Clear();
                         groups.AddRange(FindGroups());
                    } while (groups.Count > 0);
               } while (GetHint().IsEmpty);
          }

          public GameField(string file)
          {
               string[] lines = File.ReadAllLines(file).Select(itm => new string(itm.Where(char.IsDigit).ToArray())).Where(itm => !string.IsNullOrWhiteSpace(itm)).ToArray();
               if (lines.Length != Constants.FIELD_HEIGHT + Constants.FIELD_WIDTH)
                    throw new FileLoadException("Invalid number of lines!");
               for (int i = 0; i < Constants.FIELD_HEIGHT; i++)
               {
                    if (lines[i].Length != Constants.FIELD_WIDTH)
                         throw new FileLoadException("Invalid number of columns!");
                    for (int j = 0; j < Constants.FIELD_WIDTH; j++)
                         if (!char.IsDigit(lines[i][j]))
                              throw new FileLoadException($"'{lines[i][j]}' is not a valid digit!");
                         else
                         {
                              int g = int.Parse(lines[i][j].ToString());
                              if (g < 1 || g > 6)
                                   throw new FileLoadException("Numbers must be in range from 1 to 6!");
                              _field[j, i] = (Gems)g;
                         }
               }
               Gems[,] cols = null;
               for (int i = 0; i < Constants.FIELD_WIDTH; i++)
               {
                    if (cols == null)
                         cols = new Gems[Constants.FIELD_WIDTH, lines[i + Constants.FIELD_HEIGHT].Length];
                    else if (cols.GetLength(1) != lines[i + Constants.FIELD_HEIGHT].Length)
                         throw new FileLoadException("All columns must have the same number of gems!");
                    for (int j = 0; j < cols.GetLength(1); j++)
                    {
                         char c = lines[i + Constants.FIELD_HEIGHT][j];
                         if (!char.IsDigit(c))
                              throw new FileLoadException($"'{c}' is not a valid digit!");
                              int g = int.Parse(c.ToString());
                              if (g < 1 || g > 6)
                                   throw new FileLoadException("Numbers must be in range from 1 to 6!");
                              cols[i, j] = (Gems)g;
                    }
               }
               _gf = new GemFactory(cols);
          }

          public Gems this[int x, int y] => _field[x, y];

          public void RemoveGroups(GemGroup[] groups)
          {
               foreach (GemGroup g in groups)
                    foreach (Point p in g.Group)
                         _field[p.X, p.Y] = Gems.None;
               Gravity();
          }

          public void Switch(Point a, Point b)
          {
               Switch(a.X, a.Y, b.X, b.Y);
          }

          public GemGroup[] FindGroups()
          {
               List<GemGroup> ret = new List<GemGroup>();
               for (int i = 0; i < Constants.FIELD_WIDTH; i++)
                    for (int j = 0; j < Constants.FIELD_HEIGHT; j++)
                         ret.AddRange(CheckGroups(i, j));
               return ret.ToArray();
          }

          public GemGroup[] CheckGroups(int x, int y)
          {
               List<GemGroup> ret = new List<GemGroup>();
               Gems g = _field[x, y];
               if (x > 0 && x < Constants.FIELD_WIDTH - 1 && _field[x - 1, y] == g && _field[x + 1, y] == g)
                    ret.Add(new GemGroup(new Point(x - 1, y), new Point(x, y), new Point(x + 1, y)));
               if (y > 0 && y < Constants.FIELD_HEIGHT - 1 && _field[x, y - 1] == g && _field[x, y + 1] == g)
                    ret.Add(new GemGroup(new Point(x, y - 1), new Point(x, y), new Point(x, y + 1)));
               return ret.ToArray();
          }

          public Gems Peek(int c)
          {
               return _gf.Peek(c);
          }

          public Turn GetHint()
          {
               for (int i = 0; i < Constants.FIELD_WIDTH; i++)
                    for (int j = 0; j < Constants.FIELD_HEIGHT; j++)
                    {
                         Turn t = new Turn();
                         if (i > 0)
                         {
                              t = new Turn(new Point(i, j), new Point(i - 1, j));
                              if (SimulateTurn(t))
                                   return t;
                         }
                         if (j > 0)
                         {
                              t = new Turn(new Point(i, j), new Point(i, j - 1));
                              if (SimulateTurn(t))
                                   return t;
                         }
                    }
               return new Turn();
          }

          private bool SimulateTurn(Turn t)
          {
               bool ret = false;
               Switch(t.From, t.To);
               if (t.To.X - 1 >= 0 && CheckGroups(t.To.X - 1, t.To.Y).Length > 0)
                    ret = true;
               if (!ret && t.To.Y - 1 >= 0 && CheckGroups(t.To.X, t.To.Y - 1).Length > 0)
                    ret = true;
               if (!ret && t.To.X + 1 < Constants.FIELD_WIDTH && CheckGroups(t.To.X + 1, t.To.Y).Length > 0)
                    ret = true;
               if (!ret && t.To.Y + 1 < Constants.FIELD_HEIGHT && CheckGroups(t.To.X, t.To.Y + 1).Length > 0)
                    ret = true;
               Switch(t.From, t.To);
               return ret;
          }

          private void Switch(int xa, int ya, int xb, int yb)
          {
               Gems g = _field[xa, ya];
               _field[xa, ya] = _field[xb, yb];
               _field[xb, yb] = g;
          }

          private void Gravity()
          {
               for (int i = 0; i < Constants.FIELD_WIDTH; i++)
                    for (int j = 0; j < Constants.FIELD_HEIGHT; j++)
                         if (_field[i, j] == Gems.None)
                         {
                              for (int k = j; k >= 0; k--)
                              {
                                   if (k > 0)
                                        Switch(i, k, i, k - 1);
                                   else
                                        _field[i, k] = _gf.GetGem(i);
                              }
                         }
          }

          private sealed class GemFactory
          {
               private readonly Random _r;
               private readonly Gems[,] _cols;
               private readonly int[] _colIndexes;

               public GemFactory(Gems[,] g)
               {
                    _cols = g;
                    _colIndexes = new int[g.GetLength(0)];
               }

               public GemFactory()
               {
                    _r = new Random();
                    _cols = new Gems[Constants.FIELD_WIDTH, 1];
                    for (int i = 0; i < Constants.FIELD_WIDTH; i++)
                         _cols[i, 0] = (Gems)(_r.Next(6) + 1);
               }

               public Gems GetGem(int c)
               {
                    if (_r != null)
                    {
                         Gems ret = _cols[c, 0];
                         _cols[c, 0] = (Gems)(_r.Next(6) + 1);
                         return ret;
                    }
                    Gems retCol = _cols[c, _colIndexes[c]];
                    _colIndexes[c] = (_colIndexes[c] + 1) % _cols.GetLength(1);
                    return retCol;
               }

               public Gems Peek(int c)
               {
                    if (_r != null)
                         return _cols[c, 0];
                    return _cols[c, _colIndexes[c]];
               }
          }
     }
}
