﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Drawing.Drawing2D;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Windows.Forms;

namespace Jewles
{
     partial class GameWindow : Form
     {
          private readonly Game _g;
          private readonly Brush[] _brushes;
          private static readonly SolidBrush _selectionColor = new SolidBrush(Color.FromArgb(200, 0, 0, 0));
          private bool _selection = false;
          private Point _p = Point.Empty;
          private Turn _hint;

          public GameWindow(Game g)
          {
               _g = g;
               InitializeComponent();
               Size s = new Size(Constants.RENDER_STEP * Constants.FIELD_WIDTH, Constants.RENDER_STEP * (Constants.FIELD_HEIGHT + 1));
               ClientSize = new Size(s.Width, s.Height + canvas.Location.Y);
               canvas.Size = s;

               _brushes = new Brush[] {
                    new SolidBrush(Color.Black),
                    new SolidBrush(Color.Red),
                    new SolidBrush(Color.LimeGreen),
                    new SolidBrush(Color.Blue),
                    new SolidBrush(Color.Silver),
                    new SolidBrush(Color.Yellow),
                    new SolidBrush(Color.BlueViolet),
               };
               SetLabels();
          }

          protected override void OnInvalidated(InvalidateEventArgs e)
          {
               Draw();
               base.OnInvalidated(e);
          }

          protected override void OnShown(EventArgs e)
          {
               Draw();
               base.OnShown(e);
          }

          private void Draw()
          {
               SetLabels();
               Graphics g = canvas.CreateGraphics();
               g.CompositingQuality = CompositingQuality.HighQuality;
               g.Clear(Color.White);
               if (_g.Status == Game.Statuses.Playing)
               {
                    for (int i = 0; i < Constants.FIELD_WIDTH; i++)
                    {
                         g.FillRectangle(GetBrush(_g.Field.Peek(i)), i * Constants.RENDER_STEP, 0, Constants.RENDER_STEP, Constants.RENDER_STEP);
                         g.FillRectangle(new SolidBrush(Color.FromArgb(15, 255, 255, 255)), 0, 0, Constants.RENDER_STEP * Constants.FIELD_WIDTH, Constants.RENDER_STEP);
                         for (int j = 0; j < Constants.FIELD_HEIGHT; j++)
                              g.FillRectangle(GetBrush(_g.Field[i, j]), i * Constants.RENDER_STEP, (j + 1) * Constants.RENDER_STEP, Constants.RENDER_STEP, Constants.RENDER_STEP);
                    }
                    if (_selection)
                         g.FillRectangle(_selectionColor, _p.X * Constants.RENDER_STEP, (_p.Y + 1) * Constants.RENDER_STEP, Constants.RENDER_STEP, Constants.RENDER_STEP);
                    if (!_hint.IsEmpty)
                    {
                         Pen p = new Pen(new SolidBrush(Color.Black), 2);
                         g.DrawRectangle(p, _hint.From.X * Constants.RENDER_STEP, (_hint.From.Y + 1) * Constants.RENDER_STEP, Constants.RENDER_STEP - 1, Constants.RENDER_STEP - 1);
                         g.DrawRectangle(p, _hint.To.X * Constants.RENDER_STEP, (_hint.To.Y + 1) * Constants.RENDER_STEP, Constants.RENDER_STEP - 1, Constants.RENDER_STEP - 1);
                    }
               }
               else
               {
                    Font goFont = new Font(FontFamily.GenericSansSerif, 26, FontStyle.Bold);
                    Font msgFont = new Font(FontFamily.GenericSansSerif, 16, FontStyle.Italic);
                    SizeF goSize = g.MeasureString("Game Over!", goFont);
                    g.DrawString("Game Over!", goFont, new SolidBrush(Color.Black), (canvas.Width / 2) - ((int)goSize.Width / 2), (canvas.Height / 2) - ((int)goSize.Height / 2));
                    if (_g.Status == Game.Statuses.NoTurns)
                    {
                         SizeF msgSize = g.MeasureString("No turns left to play!", msgFont);
                         g.DrawString("No turns left to play!", msgFont, new SolidBrush(Color.Black), (canvas.Width / 2) - ((int)msgSize.Width / 2), (canvas.Height / 2) - ((int)msgSize.Height / 2) + goSize.Height - 10);
                    }
                    else
                    {
                         SizeF msgSize = g.MeasureString("Time is up!", msgFont);
                         g.DrawString("Time is up!", msgFont, new SolidBrush(Color.Black), (canvas.Width / 2) - ((int)msgSize.Width / 2), (canvas.Height / 2) - ((int)msgSize.Height / 2) + goSize.Height - 10);
                    }
               }
          }

          private Brush GetBrush(GameField.Gems g)
          {
               return _brushes[(byte)g];
          }

          private new void MouseClick(object sender, MouseEventArgs e)
          {
               _hint = new Turn();
               if (e.Button != MouseButtons.Left || e.Y <= Constants.RENDER_STEP)
               {
                    _selection = false;
                    Draw();
                    return;
               }
               Point p = new Point(e.Location.X / Constants.RENDER_STEP, (e.Location.Y - Constants.RENDER_STEP) / Constants.RENDER_STEP);
               if (_selection && !p.Equals(_p))
               {
                    int distX = Math.Abs(p.X - _p.X);
                    int distY = Math.Abs(p.Y - _p.Y);
                    if (((distX == 1) != (distY == 1)) && distX <= 1 && distY <= 1)
                         _g.Turn(_p, p);
                    _selection = false;
               }
               else
               {
                    _selection = true;
                    _p = p;
               }
               Draw();
          }

          private void SetLabels()
          {
               scoreLbl.Text = $"Score:{Environment.NewLine} {_g.Points}";
               if (!_g.Time)
                    timeLbl.Text = $"Remaining time:{Environment.NewLine} No time limit!";
               else if (_g.RemainingTime.ToString("g").StartsWith("-"))
                    timeLbl.Text = $"Remaining time:{Environment.NewLine} Time is up!";
               else
               timeLbl.Text = $"Remaining time:{Environment.NewLine} {_g.RemainingTime.ToString("mm\\:ss")}";

               scoreLbl.Location = new Point(10, scoreLbl.Location.Y);
               timeLbl.Location = new Point(ClientSize.Width - timeLbl.Width - 10, timeLbl.Location.Y);
          }

          private void GameWindow_KeyPress(object sender, KeyPressEventArgs e)
          {
               if (e.KeyChar == 32)
               {
                    _hint = _g.Field.GetHint();
                    Draw();
               }
          }

          private void canvas_Paint(object sender, PaintEventArgs e)
          {
               Draw();
          }

          private void timeRefresher_Tick(object sender, EventArgs e)
          {
               Game.Statuses s = _g.Status;
               _g.CheckTime();
               SetLabels();
               if (_g.Status != s)
                    Draw();
          }
     }
}
